﻿// DCC Decoder for DS-DCC decode
// By yaasan
// Based on Nicolas's sketch http://blog.nicolas.cx
// Inspired by Geoff Bunza and his 17 Function DCC Decoder & updated library
//
//
// Debug serial output available on the serial port at baud 115200, aka Tools -> Serial Monitor
//

#include <Arduino.h>
#include <string.h>
#include "motor_ctrl.h"
#include "motor_func.h"


/*************************************************
 * 
 * Declaration of Symbols
 * 
 *************************************************/

#define COUNT_ZEROMAX 10


/*************************************************
 * 
 * Declaration of Classes
 * 
 *************************************************/


/*************************************************
 * 
 * Declaration of Variables
 * 
 *************************************************/


//モータ制御関連の変数
long gMotorLPF_buf = 0;
long gBEMFLPF_buf = 0;

long gMotorPI_buf = 0;
long gMotorSpeed = 0;
byte gDirection = 128;
int gPrevSpeed = 0;
int gPWMRef = 0;
byte gZeroCount = 0;

MOTOR_PARAM gParam;

/*************************************************
 * 
 * Declaration of Functions
 * 
 *************************************************/

void MOTOR_Init();
void MOTOR_SetCV(byte iNo, byte inData);
void MOTOR_Sensor();
void MOTOR_Ack(void);
void MOTOR_Main(int inSpeedCmd, int inDirection);



/*************************************************
 * 
 * Functions
 * 
 *************************************************/

void MOTOR_Init()
{
	gParam.mStartVoltage = 32;
	gParam.mBEMFcoefficient = 3;
	gParam.mPI_P = 32;
	gParam.mPI_I = 96;
 	gParam.mBEMFcutoff = 32;
	
}

void MOTOR_SetCV(byte iNo, byte inData)
{
	switch(iNo)
	{
	case 2:
		gParam.mStartVoltage = inData;
		break;
  	case 54:
    	gParam.mBEMFcoefficient = inData;
    	break;
	case 55:
		gParam.mPI_P = inData;
		break;
	case 56:
		gParam.mPI_I = inData;
		break;
  	case 57:
    	gParam.mBEMFcutoff = inData;
    	break;

	}
	
	
}

void MOTOR_Ack(void)
{
  analogWrite(MOTOR_PWM_B, 0);
  analogWrite(MOTOR_PWM_A, 250);
  delay( 6 );  
  analogWrite(MOTOR_PWM_A, 0);
}


void MOTOR_Sensor()
{
	int aBEMF_result, aSpeed_calculated;
	
	//BEMF detection
	aBEMF_result = MOTOR_GetBEMF();
	
	
	aSpeed_calculated = aBEMF_result;
	
	if(aSpeed_calculated > 255)
	{
		aSpeed_calculated = 255;
	}
	
	//Speed conversion
	gMotorSpeed = MOTOR_LPF(aSpeed_calculated, 8, &gBEMFLPF_buf);
	
	
}


/* Motor control Task (10Hz) */
void MOTOR_Main(int inSpeedCmd, int inDirection)
{
	
	int aSpeedRef = inSpeedCmd;
	int aPWMRef = 0;
	int aSpeedRefLinear = 0;
	long aSpeedRef_offseted = 0;
	int aBEMF_result, aSpeedDet;
	
	
	//8bit LPF
	aSpeedRef = MOTOR_LPF(aSpeedRef, 64, &gMotorLPF_buf);
	
	
	//PI Control
	if( gParam.mBEMFcutoff > 0)
	{
		
		//Speed detection
//        aSpeedDet = gMotorSpeed;
        aSpeedDet = gMotorSpeed * (1- 0.1 * gParam.mBEMFcoefficient);             //MECY Adjust 2016/06/07
		
		if( aSpeedDet < gParam.mBEMFcutoff)
		{
			//ある閾値以下は速度ゼロと見なす
			aSpeedDet = 0;
		}
		else
		{
			//検出速度のオフセットをする
			aSpeedDet = aSpeedDet - gParam.mBEMFcutoff;
		}
		
		//Command zero check
		if( aSpeedDet == 0)
		{
			gZeroCount++;
			
			if( gZeroCount > COUNT_ZEROMAX)
			{
				gZeroCount = 0;
				gMotorPI_buf = 0;
			}
		}
		else
		{
			gZeroCount = 0;
		}
			
		//BEMF時 PI制御
//		aSpeedRef_offseted = MOTOR_PI(aSpeedRef - aSpeedDet, gParam.mPI_P, gParam.mPI_I, &gMotorPI_buf) + ((aSpeedRef==0) ? 0 : gParam.mStartVoltage);

/*****************************************  CV2 Impact Readjust by MECY 2016/06/28 ******************************************************/
		if((aSpeedRef==0)&&( aSpeedDet==0))
		{
		aSpeedRef_offseted = MOTOR_PI(aSpeedRef - aSpeedDet, gParam.mPI_P, gParam.mPI_I, &gMotorPI_buf) ;
		}
		else if((aSpeedRef>=1)&&( aSpeedDet>=0))
		{
		aSpeedRef_offseted = MOTOR_PI(aSpeedRef - aSpeedDet, gParam.mPI_P, gParam.mPI_I, &gMotorPI_buf) + gParam.mStartVoltage ;
		}
	}
	else
	{
		if((aSpeedRef==0)&&( aSpeedDet==0))
		{
		//PI無効
		aSpeedRef_offseted = aSpeedRef;
		}
		else if((aSpeedRef>=1)&&( aSpeedDet>=0))
		{
		aSpeedRef_offseted = aSpeedRef + gParam.mStartVoltage ; //2016/06/19 Modify Nagoden 
		}
/****************************************************************************************************************************************/              
	}

	//前回速度として保存
	gPrevSpeed = aSpeedRef_offseted ;

	//Limiter (8bit PWM Output)
	aPWMRef = MOTOR_limitSpeed((uint16_t)(aSpeedRef_offseted));
	
	//PWM出力
	if( aPWMRef == 0)
	{
		analogWrite(MOTOR_PWM_A, 0);
		analogWrite(MOTOR_PWM_B, 0);
	}
	else
	{
    //PWM出力
   		//進行方向でPWMのABを切り替える
  		  if( inDirection > 0){
	  		  analogWrite(MOTOR_PWM_B, 0);
		  	  analogWrite(MOTOR_PWM_A, aPWMRef);
		    }
		    else
		    {
		  	  analogWrite(MOTOR_PWM_A, 0);
		  	  analogWrite(MOTOR_PWM_B, aPWMRef);
		    }
	}

	gDirection = inDirection;
	gPWMRef = aPWMRef;
	
	Serial.print("SCMD= ");
	Serial.print(inSpeedCmd);

	Serial.print(" ,SREF= ");
	Serial.print(aSpeedRef);
	
	Serial.print(" ,BEMF= ");
	Serial.print(aSpeedDet);

	Serial.print(" ,PI_IBUF= ");
	Serial.print(gMotorPI_buf);

	Serial.print(" ,PWM=");
	Serial.println(aPWMRef);

	
}

int MOTOR_GetBEMF()
{
	uint32_t aAve = 0;
	int i;
	
	//Stop PWM
	analogWrite(MOTOR_PWM_B, 0);
	analogWrite(MOTOR_PWM_A, 0);
	
	delayMicroseconds(300);
	
	for( i = 0; i < 8; i++)
	{
//		aAve = aAve + analogRead(A0);
//新ＭＰ３はA1にピン変更
    aAve = aAve + analogRead(A1);
		delayMicroseconds(10);
	}
	
	//進行方向でPWMのABを切り替える
	if( gDirection > 0){
		analogWrite(MOTOR_PWM_B, 0);
		analogWrite(MOTOR_PWM_A, gPWMRef);
	}
	else
	{
		analogWrite(MOTOR_PWM_A, 0);
		analogWrite(MOTOR_PWM_B, gPWMRef);
	}
	
	return (aAve >> 3);
	
	
}



int MOTOR_limitSpeed(int inSpeed)
{
//	uint16_t aSpeed_ref = 0;
	uint16_t aSpeedref = 0;                                   //MECY Adjust 2016/06/04
	
	if( inSpeed >= MOTOR_LIM_MAX)
	{
//		aSpeed_ref = MOTOR_LIM_MAX;
		aSpeedref = MOTOR_LIM_MAX;                            //MECY Adjust 2016/06/04
	}
	else if( inSpeed <= 0)
	{
//		aSpeed_ref = 0;
		aSpeedref = 0;                                        //MECY Adjust 2016/06/04
	}
//	else
	else if( gParam.mBEMFcutoff == 0 )                        //MECY Adjust 2016/06/04
	{
		aSpeedref = inSpeed;
	}
	else if( gParam.mBEMFcutoff > 0 )
	{
		aSpeedref = inSpeed;
	}
	
//	return aSpeed_ref;                                        //MECY Adjust 2016/06/04
	return aSpeedref;
}

